# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import sys
from pathlib import Path
from typing import Optional

import click
import jinja2


def get_venv_share_directory(app_name: str) -> Optional[Path]:
    '''
    Find the share directory within the current virtual environment.

    Args:
        app_name: The name of your application

    Returns:
        Path object to the virtual environment share directory, or None if not
        in a venv
    '''
    # Check if running in a virtual environment
    if sys.prefix != sys.base_prefix:
        # Get the virtual environment root
        venv_root = Path(sys.prefix)

        # Common share directory locations in a venv
        possible_share_dirs = [
            venv_root / 'share',            # /venv/share
            venv_root / 'local' / 'share',  # /venv/local/share
            venv_root / 'var' / 'share',    # /venv/var/share
        ]

        # Check app-specific directories first
        for base_dir in possible_share_dirs:
            app_dir = base_dir / app_name
            if app_dir.exists():
                return app_dir

        # If app-specific directory doesn't exist, return first existing share directory
        for base_dir in possible_share_dirs:
            if base_dir.exists():
                return base_dir / app_name

    return None


def share_directory_list(app_name):
    '''
    A list of possible share locations.

    Args:
        app_name: The name of your application

    Returns:
        A list of Path objects pointing to the share directories.
    '''
    # First check virtual environment (includes pipx environments)
    venv_share = get_venv_share_directory(app_name)
    if venv_share and venv_share.exists():
        yield venv_share

    # Get XDG_DATA_HOME from environment or use default
    xdg_data_home = os.environ.get('XDG_DATA_HOME', '~/.local/share')
    xdg_data_home = Path(xdg_data_home).expanduser()

    # Get XDG_DATA_DIRS from environment or use default
    xdg_data_dirs = os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share')
    data_dirs = [Path(p) for p in xdg_data_dirs.split(':')]

    # Order of preference for share directories
    search_paths = [
        xdg_data_home / app_name,  # ~/.local/share/<app_name>
        *[p / app_name for p in data_dirs]  # System-wide directories
    ]

    # Return the first directory that exists and contains data
    for path in search_paths:
        if path.exists():
            yield path


def filename_list(app_name, filename):
    # First test if the file name supplied exists
    yield filename
    # If unsuccessful try to open the file from other places
    for path in share_directory_list(app_name):
        yield path / 'templates' / filename


def load_template(app_name, filename):
    '''
    Loads a template stored in `filename`
    '''
    for fn in filename_list(app_name, filename):
        try:
            with open(fn, 'r') as f:
                template_str = f.read()
            return template_str
        except FileNotFoundError:
            pass
    raise click.UsageError(f'File "{filename}" not found')


def render(app_name, filename, context):
    template_str = load_template(app_name, filename)
    template = jinja2.Template(template_str)
    return template.render(**context)



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
caldavctl uses a simple format to create new objects. this format is parsed and
converted into iCalendar objects.
'''

from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from datetime import datetime
from abc import ABC, abstractmethod


STRING = 0
LIST = 1
DATE = 2
INTEGER = 3
TZ = 4

# States:
KEY = 0
VALUE = 1
COMMENT = 2


class ValueParser(ABC):
    def __init__(self, value):
        self.value = value
        self.tokens = ()

    @abstractmethod
    def tokenize(self):
        ...

    @abstractmethod
    def validate(self):
        ...

    def parse(self):
        self.tokenize()
        self.validate()
        return self.tokens


class ParserError(Exception):
    pass


class ObjectParser:
    def __init__(self, object, lexicon):
        self.object = object
        self.lexicon = lexicon

    def tokenize(self):
        data = self.object
        tokens = []
        pos = 0
        lenght = len(data)
        state = KEY
        while pos < lenght:
            if data[pos] in ' \t':  # Ignore spaces
                pos += 1
                continue
            if data[pos] == '#':  # Ignore comments
                pos = data.find('\n', pos) + 1
                if pos == 0:  # End of file
                    pos = lenght
                continue

            if state == KEY:  # Find the key
                if data[pos] == '\n':
                    pos += 1
                    continue
                end = data.find(':', pos)
                if end == -1:
                    raise ParserError('Could not find key/value pair.')
                key = data[pos:end].strip().lower()
                pos = end + 1
                state = VALUE
                continue

            if state == VALUE:  # Find the value
                if pos == lenght or data[pos] == '\n':  # End of file or empty value
                    pos += 1
                    state = KEY
                    continue
                if data[pos:pos + 2] == '[[':  # Multi line string
                    end = data.find(']]', pos)
                    value = data[pos + 2:end].strip()
                    tokens.append([key, value])
                    pos = end + 2
                    state = KEY
                    continue
                else:
                    # Single line string
                    end = data.find('\n', pos)
                    if end == -1:  # End of file
                        end = lenght
                    comment = data.find('#', pos)
                    if comment < end and comment != -1:
                        value = data[pos:comment].strip()
                    else:
                        value = data[pos:end].strip()
                    tokens.append([key, value])
                    pos = end + 1
                    state = KEY
                    continue
            pos += 1
        return tokens

    def parse(self, tokens):
        result = {}
        for key, value in tokens:
            if key not in self.lexicon:
                raise ParserError(f'Unknown key: "{key}".')

            optional = self.lexicon[key][1]
            value_type = self.lexicon[key][0]

            if not value and not optional:
                raise ParserError(f'Key "{key}" is not optional')
            elif not value and optional:
                continue

            if value_type == STRING:
                result[key] = value
            elif value_type == LIST:
                result[key] = [v.strip() for v in value.split(',')]
            elif value_type == DATE:
                try:
                    result[key] = datetime.fromisoformat(value)
                except ValueError:
                    raise ParserError(f'Invalid date format for: "{value}"')
            elif value_type == INTEGER:
                try:
                    result[key] = int(value)
                except ValueError:
                    raise ParserError(f'Invalid integer: "{value}"')
            elif value_type == TZ:
                try:
                    result[key] = ZoneInfo(value)
                except ZoneInfoNotFoundError:
                    raise ParserError(f'Invalid timezone: "{value}"')
            elif issubclass(value_type, ValueParser):
                if key not in result:
                    result[key] = []
                result[key].append(value_type(value).parse())

        return result

    def run(self):
        tokens = self.tokenize()
        result = self.parse(tokens)
        return result



import pytest
from datetime import datetime, timezone
from zoneinfo import ZoneInfo
from caldavctl.object_parser import ObjectParser, ParserError
from caldavctl.event_builder import parse_event, check_if_naive, EVENT_LEXICON


sample_event_data_i = """
        SUMMARY: This is a description
        LOCATION: [[Timbuktu]] # With a comment
        CATEGORIES: Birthday, Shopping, Training
        TIMEZONE: Europe/Lisbon
        DTSTART: 2024-01-20 09:00
        DTEND: 2024-01-20 09:45
        DESCRIPTION: [[
        Once upon a time,
        There was a little lamb
        ]]
        """

sample_event_data_ii = '''# Mandatory
SUMMARY: Example summary
DTSTART: 2024-12-30 09:00
DTEND: 2024-12-30 09:45

# Optional
LOCATION:
CATEGORIES:
TIMEZONE: Europe/Lisbon
DESCRIPTION: [[ ]]

# NOTES:
#
#   * Date and time:
#       * The dates must be in iso format, for instance: 2024-12-29 13:45;
#       * The timezone used if the one defined by default or the one defined in
#         "TIMEZONE";
#   * Categories: the categories are a comma separeted list;
#   * Description: The description can be multi line, just make sure it's
#     delimited by [[ ]].'''


@pytest.fixture
def default_timezone():
    return ZoneInfo('UTC')


def test_tokenize():
    parser = ObjectParser(sample_event_data_i, EVENT_LEXICON)
    tokens = parser.tokenize()
    expected_tokens = [
        ['summary', 'This is a description'],
        ['location', 'Timbuktu'],
        ['categories', 'Birthday, Shopping, Training'],
        ['timezone', "Europe/Lisbon"],
        ['dtstart', '2024-01-20 09:00'],
        ['dtend', '2024-01-20 09:45'],
        ['description', 'Once upon a time,\n        There was a little lamb']
    ]

    assert tokens == expected_tokens


def test_parse_i():
    parser = ObjectParser(sample_event_data_i, EVENT_LEXICON)
    tokens = parser.tokenize()
    result = parser.parse(tokens)

    expected_result = {
        'summary': 'This is a description',
        'location': 'Timbuktu',
        'categories': ['Birthday', 'Shopping', 'Training'],
        'timezone': ZoneInfo('Europe/Lisbon'),
        'dtstart': datetime(2024, 1, 20, 9, 0),
        'dtend': datetime(2024, 1, 20, 9, 45),
        'description': 'Once upon a time,\n        There was a little lamb'
    }

    assert result['summary'] == expected_result['summary']
    assert result['location'] == expected_result['location']
    assert result['categories'] == expected_result['categories']
    assert result['description'] == expected_result['description']
    assert result['dtstart'] == expected_result['dtstart']
    assert result['dtend'] == expected_result['dtend']
    assert result['timezone'] == expected_result['timezone']


def test_parse_ii():
    parser = ObjectParser(sample_event_data_ii, EVENT_LEXICON)
    tokens = parser.tokenize()
    result = parser.parse(tokens)

    expected_result = {
        'summary': 'Example summary',
        'location': '',
        'categories': [],
        'timezone': ZoneInfo('Europe/Lisbon'),
        'dtstart': datetime(2024, 12, 30, 9, 0),
        'dtend': datetime(2024, 12, 30, 9, 45),
        'description': ''
    }

    assert result['summary'] == expected_result['summary']
    assert 'location' not in result
    assert 'categories' not in result
    assert 'description' not in result
    assert result['dtstart'] == expected_result['dtstart']
    assert result['dtend'] == expected_result['dtend']
    assert result['timezone'] == expected_result['timezone']


def test_check_if_naive():
    naive_date = datetime(2024, 1, 20, 9, 0)
    tz = ZoneInfo('Europe/Lisbon')
    localized_date = check_if_naive(naive_date, tz)

    assert f'{localized_date.tzinfo}' == 'Europe/Lisbon'
    assert localized_date.utcoffset() == tz.utcoffset(naive_date)


def test_parse_event_with_naive_dates(default_timezone):
    event_with_naive_dates = (
        """
        SUMMARY: Test Event
        LOCATION: Timbuktu
        TIMEZONE: Europe/Lisbon
        DTSTART: 2024-01-20 09:00
        DTEND: 2024-01-20 09:45Z  # UTC date
        """
    )

    result = parse_event(event_with_naive_dates, default_timezone)
    tz = ZoneInfo('Europe/Lisbon')

    assert result['dtstart'].tzinfo == tz
    assert result['dtend'].tzinfo == timezone.utc


def test_unknown_key():
    event_with_unknown_key = (
        """
        SUMMARY: Test Event
        UNKNOWN_KEY: This should fail
        """
    )
    parser = ObjectParser(event_with_unknown_key, EVENT_LEXICON)

    with pytest.raises(ParserError, match='Unknown key:'):
        parser.run()


def test_invalid_date():
    event_with_invalid_date = (
        """
        SUMMARY: Test Event
        DTSTART: invalid-date
        """
    )

    parser = ObjectParser(event_with_invalid_date, EVENT_LEXICON)

    with pytest.raises(ParserError, match='Invalid date format'):
        parser.run()


def test_event_no_timezone():
    sample_event_no_timezone = '''# Mandatory
    SUMMARY: Example summary
    DTSTART: 2024-12-30 09:00
    DTEND: 2024-12-30 09:45

    # Optional
    LOCATION:
    CATEGORIES:
    DESCRIPTION: [[ ]]'''

    result = parse_event(sample_event_no_timezone, ZoneInfo('Europe/Lisbon'))

    assert f'{result['dtstart'].tzinfo}' == 'Europe/Lisbon'
    assert f'{result['dtend'].tzinfo}' == 'Europe/Lisbon'



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from datetime import datetime, timedelta

import click
from caldav.lib import error

from caldavctl import dav, get_name
from caldavctl.click_utils import create_compatibility_check
from caldavctl.event_builder import event_builder, parse_event
from caldavctl.object_parser import ParserError
from caldavctl.templates import render
from caldavctl.utils import edit_text_with_editor, to_datetime

# Configuration

EMPTY_EVENT = '''# Mandatory
DTSTART: %(start)s
DTEND: %(end)s

# Recommended
SUMMARY:

# Optional
LOCATION:
CATEGORIES:
TIMEZONE:
PRIORITY: 5
ALARM: P0D
DESCRIPTION: [[ ]]

# NOTES:
#
#   * Date and time:
#       * The dates must be in iso format, for instance: 2024-12-29 13:45;
#       * The timezone used is the one defined by default or the one defined in
#         the "TIMEZONE" key;
#   * Categories: the categories are a comma separated list;
#   * Priority: 1 - highest priority, 9 - lowest priority
#   * Description: The description can be multi line, just make sure it's
#     delimited by [[ ]].
#   * Alarms:
#       * By default an alarm is defined the moment the event begins;
#       * The alarms are defined using:
#           ALARM: <trigger> [<description>]
#       * The trigger is defined according to:
#           https://datatracker.ietf.org/doc/html/rfc5545#autoid-38
#       * For example, to trigger an alarm 30 minutes before the event start, use:
#           ALARM: -PT30M
#       * The alarm description is optional
'''


# Commands:

exclusive_option = create_compatibility_check(
    ['today', 'week', 'dtstart', 'dtend', 'day'],  # Exclusive options
    [['dtstart', 'dtend']])  # Except these two that can be used together


@click.command('list', options_metavar='[options]')
@click.option('-t', '--today',
              is_flag=True, show_default=True, default=False,
              callback=exclusive_option, help='Today\'s events',)
@click.option('-w', '--week',
              is_flag=True, show_default=True, default=False,
              callback=exclusive_option, help='This week\'s events')
@click.option('-s', '--dtstart',
              callback=exclusive_option, help='Date range, start date', metavar='<yyyy-mm-dd>')
@click.option('-e', '--dtend',
              callback=exclusive_option, help='Date range, end date', metavar='<yyyy-mm-dd>')
@click.option('-d', '--day',
              callback=exclusive_option, help='Events on the day', metavar='<yyyy-mm-dd>')
@click.option('-sd', '--show-description',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s description.')
@click.option('-si', '--show-uid',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s UID.')
@click.option('-tf', '--template-file', default='event.txt',
              help='Template used to format the output', metavar='<file>')
@click.pass_obj
def list_events(context, today, week, dtstart, dtend, day, show_description, show_uid, template_file):
    '''
    List events from the server list.
    '''
    tz = context['config'].tz()
    start_date = None
    end_date = None

    if today:
        start_date = datetime.now().replace(hour=0, minute=0, second=0, tzinfo=tz)
        end_date = datetime.now().replace(hour=23, minute=59, second=59, tzinfo=tz)
    elif week:
        today_date = datetime.now().replace(hour=0, minute=0, second=0, tzinfo=tz)
        start_date = today_date - timedelta(days=today_date.weekday())
        end_date = start_date + timedelta(days=6)
        end_date = end_date.replace(hour=23, minute=59, second=59)
    elif dtstart and dtend:
        if dtstart:
            start_date = to_datetime(dtstart, tz)
        else:
            start_date = datetime(1900, 1, 1, tzinfo=tz)
        if dtend:
            end_date = to_datetime(dtend, tz)
        else:
            end_date = datetime(5000, 1, 1, tzinfo=tz)
    elif day:
        try:
            dt = datetime.fromisoformat(day)
            start_date = dt.replace(hour=0, minute=0, second=0, tzinfo=tz)
            end_date = dt.replace(hour=23, minute=59, second=59, tzinfo=tz)
        except ValueError:
            raise click.UsageError(f'Invalid date "{day}"')
    else:
        raise click.UsageError('You have to specify a time frame, for example with -t or -w.')

    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        # Check if the calendar supports events:
        if 'VEVENT' not in calendar.get_supported_components():
            raise click.UsageError('This calendar does not support events.')

        # Get events in time range
        events = calendar.search(
            start=start_date,
            end=end_date,
            event=True,
            expand=True,
            sort_keys=['dtstart']
        )
        event_list = []
        for event in events:
            ev = event.icalendar_component
            date = ev.get('dtstart').dt
            event_context = {
                'date': date.strftime('%Y-%m-%d %H:%M') if isinstance(date, datetime) else f'{date} (day)',
                'summary': ev.get('summary', ''),
                'description': ev.get('description', ''),
                'uid': ev.get('uid')
            }
            event_list.append(event_context)

        template_context = {
            'show_uid': show_uid,
            'show_description': show_description,
            'event_list': event_list,
            'newline': '\n'
        }

        txt = render(get_name(), template_file, template_context)
        click.echo(txt.strip())


@click.command('create', options_metavar='[options]')
@click.option('-f', '--file', type=click.File('r'), default=sys.stdin, help='Create event from this file', metavar='<file>')
@click.option('-e', '--edit', is_flag=True, default=False, help='Edit the event on a text editor?')
@click.option('--dry-run', is_flag=True, default=False, help='Show only the iCalendar generated')
@click.pass_obj
def create_event(context, file, edit, dry_run):
    '''
    Create new event

    The event data can be read from stdin or from a file. Optionally the event
    can be edited using the default $EDITOR. If the option to edit is enabled
    and no file is defined using -f/--file, then an empty event is opened.
    '''
    tz = context['config'].tz()
    # Read the event definition
    if edit and file is sys.stdin:
        tomorrow = datetime.now() + timedelta(days=1)
        event_string = EMPTY_EVENT % {
            'tz': f'{tz}',
            'start': tomorrow.replace(hour=9, minute=0, second=0).strftime('%Y-%m-%d %H:%M'),
            'end': tomorrow.replace(hour=9, minute=45, second=0).strftime('%Y-%m-%d %H:%M'),
        }
    else:
        event_string = file.read()

    if edit:
        # Edit the file in $VISUAL or $EDITOR
        old_string = event_string
        event_string = edit_text_with_editor(event_string, suffix='.calendar_event')
        if old_string == event_string:
            click.confirm('No changes made. Do you want to continue?', abort=True)

    try:
        event_data = parse_event(event_string, tz)
    except ParserError as msg:
        raise click.UsageError(msg)

    event_ics = event_builder(event_data, tz)

    if dry_run:
        click.echo(event_ics)
    else:
        # Save the event to the calendar
        _, server = context['config'].get_server()
        calendar_id = context['config'].get_calendar()
        with dav.caldav_calendar(server, calendar_id) as calendar:
            # Check if the calendar supports events:
            if 'VEVENT' not in calendar.get_supported_components():
                raise click.UsageError('This calendar does not support events.')
            try:
                calendar.save_event(event_ics)
            except error.AuthorizationError as msg:
                raise click.UsageError(f'Error saving the event to the server: {msg}')

        click.echo('Event created successfully')


@click.command('delete', options_metavar='[options]')
@click.argument('event_id', metavar='<uid>')
@click.pass_obj
def delete_event(context, event_id):
    '''Delete an event on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_event(server, calendar_id, event_id) as event:
        event.delete()
    print('Event deleted')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from zoneinfo import ZoneInfo

import pytest

import click

from unittest.mock import patch, mock_open
from caldavctl.config import Config


EMPTY_CONFIG = '''
option = []

[default]
calendar = "default_calendar"
server = "default_server"
timezone = "UTC"

[server.default]
username = "default_user"
password = "default_pass"
url = "http://default.url"
'''


@pytest.fixture
def mock_user_config_dir(tmp_path):
    """Mock the user configuration directory."""
    config_dir = tmp_path / 'caldavctl'
    config_dir.mkdir()
    return config_dir


@pytest.fixture
def default_config_file(mock_user_config_dir):
    """Create a default configuration file."""
    config_file = mock_user_config_dir / 'config.toml'
    with open(config_file, 'w') as f:
        f.write(EMPTY_CONFIG)
    return config_file


@patch('caldavctl.config.user_config_dir')
@patch('os.stat')
def test_load_config_file(mock_stat, mock_user_config_dir_func, default_config_file):
    mock_user_config_dir_func.return_value = str(default_config_file.parent)
    mock_stat.return_value.st_mode = 0o600  # Valid permissions

    config = Config(None)

    assert config.config['default']['calendar'] == 'default_calendar'
    assert config.config['default']['server'] == 'default_server'
    assert config.config['default']['timezone'] == ZoneInfo('UTC')
    assert 'default' in config.config['server']
    assert config.config['server']['default']['username'] == 'default_user'
    assert config.config['server']['default']['password'] == 'default_pass'
    assert config.config['server']['default']['url'] == 'http://default.url'


@patch('os.stat')
def test_check_file_permission(mock_stat, default_config_file):
    mock_stat.return_value.st_mode = 0o600  # Valid permissions
    config = Config(default_config_file)
    assert config.config is not None

    mock_stat.return_value.st_mode = 0o644  # Invalid permissions
    with pytest.raises(click.UsageError,
                       match='The configuration file must be readable only by the user.'):
        config.check_file_permission(str(default_config_file))


@patch('builtins.open', new_callable=mock_open, read_data=b'invalid toml')
@patch('os.stat')
def test_load_config_file_invalid_toml(mock_stat, mock_file):
    mock_stat.return_value.st_mode = 0o600  # Valid permissions
    with pytest.raises(click.UsageError, match='Invalid configuration file'):
        Config('dummy_config_file.toml')


@patch('os.path.exists', return_value=False)
def test_load_config_file_not_found(mock_exists):
    config = Config('dummy_config_file.toml')
    assert config.config['default']['calendar'] is None
    assert config.config['default']['server'] is None
    assert config.config['default']['timezone'] == ZoneInfo('UTC')


@patch('os.stat')
@patch('caldavctl.config.user_config_dir')
def test_merge_server(mock_user_config_dir_func, mock_stat, default_config_file):
    mock_user_config_dir_func.return_value = str(default_config_file.parent)
    mock_stat.return_value.st_mode = 0o600  # Valid permissions
    config = Config(None, name='new_server', username='new_user',
                    password='new_pass', url='http://new.url')

    assert 'new_server' in config.config['server']
    assert config.config['server']['new_server']['username'] == 'new_user'
    assert config.config['server']['new_server']['password'] == 'new_pass'
    assert config.config['server']['new_server']['url'] == 'http://new.url'


@patch('os.stat')
@patch('caldavctl.config.user_config_dir')
def test_merge_defaults(mock_user_config_dir_func, mock_stat, default_config_file):
    mock_user_config_dir_func.return_value = str(default_config_file.parent)
    mock_stat.return_value.st_mode = 0o600  # Valid permissions
    config = Config(None, default_server='new_default_server',
                    default_calendar='new_default_calendar',
                    timezone='Europe/Lisbon')

    assert config.config['default']['server'] == 'new_default_server'
    assert config.config['default']['calendar'] == 'new_default_calendar'
    assert config.config['default']['timezone'] == ZoneInfo('Europe/Lisbon')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .__about__ import __version__


def get_version():
    return __version__


def get_name():
    return __name__



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
CalDAV tool
'''

import click

import caldavctl.calendar_dav as calendar
import caldavctl.event as event
import caldavctl.todo as todo
import caldavctl.journal as journal
import caldavctl.ics as ics
import caldavctl.backup as backup

from caldavctl import get_version
from caldavctl.config import Config


def print_version(context, param, value):
    # See https://click.palletsprojects.com/en/stable/options/#callbacks-and-eager-options
    if not value or context.resilient_parsing:
        return
    click.echo(f'caldavctl v{get_version()}')
    context.exit(0)


# General options

@click.group(options_metavar='[options]')
@click.option('-c', '--config', 'config_file', envvar='CALDAV_CONFIG', help='Configuration file', metavar='<file>')
@click.option('--name', envvar='CALDAV_NAME', help='Server nickname', metavar='<nickname>')
@click.option('--username', envvar='CALDAV_USERNAME', help='Username on the CalDAV server', metavar='<username>')
@click.option('--passwd', envvar='CALDAV_PASSWD', help='Password on the CalDAV server', metavar='<passwd>')
@click.option('--url', envvar='CALDAV_URL', help='Calendar CalDAV url', metavar='<url>')
@click.option('--timezone', envvar='CALDAV_TIMEZONE', help='Your time zone', metavar='<timezone>')
@click.option('--server', help='Default server (use nickname)', metavar='<nickname>')
@click.option('--calendar', help='Default calendar id', metavar='<uid>')
@click.option('--version', is_flag=True, is_eager=True, expose_value=False, callback=print_version,
              help='caldavctl version')
@click.pass_context
def cli_group(context,
              config_file,
              name, username, passwd, url, timezone,
              server, calendar):
    '''
    caldavctl - command line CalDAV client
    '''

    config = Config(
        config_file,
        name,
        username,
        passwd,
        url,
        timezone,
        server,
        calendar
    )

    context.obj = {
        'config': config,
        'option': []
    }


# CALENDAR

@cli_group.group('calendar', options_metavar='[options]')
@click.pass_context
def calendar_commands(context):
    '''Commands that deal with the calendars on the server'''
    pass


calendar_commands.add_command(calendar.list_calendars)
calendar_commands.add_command(calendar.create_calendar)
calendar_commands.add_command(calendar.delete_calendar)


# EVENTS

@cli_group.group('event', options_metavar='[options]')
@click.pass_context
def event_commands(context):
    '''Event management'''
    pass


event_commands.add_command(event.list_events)
event_commands.add_command(event.create_event)
event_commands.add_command(event.delete_event)


# TODOS

@cli_group.group('todo', options_metavar='[options]')
@click.pass_context
def todo_commands(context):
    '''Todo management'''
    pass


todo_commands.add_command(todo.list_todos)
todo_commands.add_command(todo.create_todo)
todo_commands.add_command(todo.delete_todo)
todo_commands.add_command(todo.toggle_todo_complete)
todo_commands.add_command(todo.percentage_complete)


# JOURNALS

@cli_group.group('journal', options_metavar='[options]')
@click.pass_context
def journal_commands(context):
    '''Journal management'''
    pass


journal_commands.add_command(journal.list_journals)


# ICS

@cli_group.group('ics', options_metavar='[options]')
@click.pass_context
def ics_commands(context):
    '''iCalendar file operations'''
    pass


ics_commands.add_command(ics.get_todo)
ics_commands.add_command(ics.create_todo)
ics_commands.add_command(ics.get_event)


# Backup

@cli_group.group('br', options_metavar='[options]')
@click.pass_context
def backup_and_restore_commands(context):
    '''Backup or restore a calendar'''
    pass


backup_and_restore_commands.add_command(backup.backup_calendar)
backup_and_restore_commands.add_command(backup.restore_calendar)


def main():
    cli_group()



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import timedelta

import pytest

from caldavctl.utils import deep_merge_dict, duration_to_timedelta, duration_validation


def test_deep_merge_dict_simple():
    source = {'a': 1, 'b': 2}
    destination = {'b': 3, 'c': 4}
    result = deep_merge_dict(source, destination)
    expected = {'a': 1, 'b': 2, 'c': 4}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


def test_deep_merge_dict_nested():
    source = {'a': {'x': 1}, 'b': 2}
    destination = {'a': {'y': 3}, 'c': 4}
    result = deep_merge_dict(source, destination)
    expected = {'a': {'x': 1, 'y': 3}, 'b': 2, 'c': 4}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


def test_deep_merge_dict_overwrite():
    source = {'a': {'x': 5}, 'b': 6}
    destination = {'a': {'x': 1, 'y': 3}, 'c': 4}
    result = deep_merge_dict(source, destination)
    expected = {'a': {'x': 5, 'y': 3}, 'b': 6, 'c': 4}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


def test_deep_merge_dict_empty_source():
    source = {}
    destination = {'a': 1, 'b': 2}
    result = deep_merge_dict(source, destination)
    expected = {'a': 1, 'b': 2}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


def test_deep_merge_dict_empty_destination():
    source = {'a': {'x': 1}, 'b': 2}
    destination = {}
    result = deep_merge_dict(source, destination)
    expected = {'a': {'x': 1}, 'b': 2}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


def test_deep_merge_dict_conflicting_types():
    source = {'a': {'x': 1}}
    destination = {'a': 5}
    result = deep_merge_dict(source, destination)
    expected = {'a': {'x': 1}}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


def test_deep_merge_dict_nested_empty_dict():
    source = {'a': {}}
    destination = {'a': {'x': 1}}
    result = deep_merge_dict(source, destination)
    expected = {'a': {'x': 1}}
    assert result == expected
    assert destination == expected  # Ensure in-place modification


test_data_duration_validation = (
    ("P1W", True),         # 1 week
    ("P1D", True),         # 1 day
    ("PT1H", True),        # 1 hour
    ("PT1M", True),        # 1 minute
    ("PT1S", True),        # 1 second
    ("P1DT1H", True),      # 1 day, 1 hour
    ("P1DT1M", True),      # 1 day, 1 minute
    ("PT1H1M", True),      # 1 hour, 1 minute
    ("PT1H1S", True),      # 1 hour, 1 second
    ("PT1M1S", True),      # 1 minute, 1 second
    ("P1DT1H1M1S", True),  # 1 day, 1 hour, 1 minute, 1 second
    ("-P1W", True),        # Negative 1 week
    ("+P1W", True),        # Positive 1 week
    ("P", False),          # Empty duration
    ("PT", False),         # Empty time duration
    ("P1H", False),        # Missing T for time
    ("P1S", False),        # Missing T for time
    ("PPT1H", False),      # Double P
    ("P1DT", False),       # Empty time part
    ("P1WT1H", False),     # Weeks cannot be combined with other units
    ("PT1D", False),       # Days must not appear after T
)


@pytest.mark.parametrize('duration, match', test_data_duration_validation)
def test_duration_validation(duration: str, match: bool):
    result = duration_validation(duration)
    print(f"{duration} {'passes' if match == result else 'does not validade'}")
    assert result == match


test_data = (
    # Basic positive durations
    ('P1W', timedelta(weeks=1)),
    ('P1D', timedelta(days=1)),
    ('PT1H', timedelta(hours=1)),
    ('PT1M', timedelta(minutes=1)),
    ('PT1S', timedelta(seconds=1)),

    # Basic negative durations
    ('-P1W', -timedelta(weeks=1)),
    ('-P1D', -timedelta(days=1)),
    ('-PT1H', -timedelta(hours=1)),
    ('-PT1M', -timedelta(minutes=1)),
    ('-PT1S', -timedelta(seconds=1)),

    # Complex positive durations
    ('P1DT2H', timedelta(days=1, hours=2)),
    ('P1DT2H3M', timedelta(days=1, hours=2, minutes=3)),
    ('P1DT2H3M4S', timedelta(days=1, hours=2, minutes=3, seconds=4)),
    ('PT2H3M4S', timedelta(hours=2, minutes=3, seconds=4)),
    ('PT3M4S', timedelta(minutes=3, seconds=4)),

    # Complex negative durations
    ('-P1DT2H', -timedelta(days=1, hours=2)),
    ('-P1DT2H3M', -timedelta(days=1, hours=2, minutes=3)),
    ('-P1DT2H3M4S', -timedelta(days=1, hours=2, minutes=3, seconds=4)),
    ('-PT2H3M4S', -timedelta(hours=2, minutes=3, seconds=4)),
    ('-PT3M4S', -timedelta(minutes=3, seconds=4)),

    # Large numbers
    ('P365D', timedelta(days=365)),
    ('PT24H', timedelta(hours=24)),
    ('PT60M', timedelta(minutes=60)),
    ('PT3600S', timedelta(seconds=3600)),

    # Common real-world cases
    ('PT15M', timedelta(minutes=15)),     # Common meeting reminder
    ('P1DT12H', timedelta(days=1, hours=12)),  # Day and a half
    ('PT30M', timedelta(minutes=30)),     # Half hour
    ('-PT15M', -timedelta(minutes=15)),   # 15 minutes before
    ('P7D', timedelta(weeks=1)),          # One week expressed in days

    # Edge cases
    ('PT0S', timedelta(seconds=0)),       # Zero duration
    ('P0D', timedelta(days=0)),           # Zero days
    ('PT1H0M0S', timedelta(hours=1)),     # Explicit zeros
    ('P1DT0H0M0S', timedelta(days=1)),    # Explicit zeros

    # Mixed cases with weird spacing (if your function handles it)
    ('P1DT23H45M10S', timedelta(days=1, hours=23, minutes=45, seconds=10)),
    ('-P1DT23H45M10S', -timedelta(days=1, hours=23, minutes=45, seconds=10)),
    (' P1DT23H45M10S ', timedelta(days=1, hours=23, minutes=45, seconds=10)),  # Extra spaces

    # Weeks
    ('P2W', timedelta(weeks=2)),
    ('-P2W', -timedelta(weeks=2)),
)


@pytest.mark.parametrize('duration, result', test_data)
def test_duration_to_timedelta(duration, result):
    td = duration_to_timedelta(duration)
    print(f'Testing "{duration}" computed {td} expected result {result} - {"passed" if td == result else "failed"}')
    assert td == result



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
import uuid

import click

from caldavctl import dav, get_name
from caldavctl.templates import render

from icalendar import Todo


@click.command('list', options_metavar='[options]')
@click.option('-sd', '--show-description', is_flag=True, show_default=True, default=False,
              help='Show the todo\'s description.')
@click.option('-si', '--show-uid', is_flag=True, show_default=True, default=False,
              help='Show the todo\'s UID.')
@click.option('-a', '--all', is_flag=True, show_default=True, default=False,
              help='Show all todos, including completed todos.')
@click.option('-tf', '--template-file', default='todo.txt',
              help='Template used to format the output', metavar='<file>')
@click.pass_obj
def list_todos(context, show_description, show_uid, all, template_file):
    '''
    List todos from the default server and default calendar
    '''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        # Check if the calendar supports events:
        if 'VTODO' not in calendar.get_supported_components():
            raise click.UsageError(f'This calendar "{calendar.name}" does not support todos.')
        # Get events in time range
        todos = calendar.todos(include_completed=all)
        todo_list = []
        for todo in todos:
            td = todo.icalendar_component
            todo_context = {
                'summary': td.get('summary'),
                'status': '[✓]' if td.get('status', '') else '[ ]',
                'percent': td.get('percent-complete', 0),
                'description': td.get('description', ''),
                'uid': td.get('uid'),
            }
            todo_list.append(todo_context)

        template_context = {
            'show_uid': show_uid,
            'show_description': show_description,
            'todo_list': todo_list,
            'newline': '\n'
        }

        txt = render(get_name(), template_file, template_context)
        click.echo(txt.strip())


@click.command('create', options_metavar='[options]')
@click.argument('summary', metavar='<summary>')
@click.option('--description', metavar='<description>')
@click.option('--due-date', metavar='<yyyy-mm-dd>')
@click.option('--priority', metavar='<int>')
@click.pass_obj
def create_todo(context, summary, description, due_date, priority):
    '''Create a todo on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    tz = context['config'].tz()

    with dav.caldav_calendar(server, calendar_id) as calendar:
        todo = Todo()

        todo.add('uid', str(uuid.uuid4()))
        todo.add('summary', summary)
        if description:
            todo.add('description', description)
        if due_date:
            try:
                dt = datetime.fromisoformat(due_date).replace(tzinfo=tz)
            except ValueError:
                raise click.UsageError(f'Invalid value for due date "{due_date}"')
            todo.add('due', dt)
        if priority:
            todo.add('priority', priority)
        todo.add('created', datetime.now(tz))
        todo.add('last-modified', datetime.now(tz))

        # Convert the todo to a string
        todo_string = todo.to_ical().decode('utf-8')

        try:
            calendar.save_event(todo_string)
            click.echo(f"todo created successfully: {summary}")
        except Exception as e:
            click.echo(f"Error creating todo: {e}")


@click.command('delete', options_metavar='[options]')
@click.argument('todo_id', metavar='<uid>')
@click.pass_obj
def delete_todo(context, todo_id):
    '''Delete a todo on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_todo(server, calendar_id, todo_id) as todo:
        todo.delete()
    click.echo('Todo deleted')


@click.command('toggle', options_metavar='[options]')
@click.argument('todo_id', metavar='<uid>')
@click.pass_obj
def toggle_todo_complete(context, todo_id):
    '''Toggle todo completed'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    tz = context['config'].tz()

    with dav.caldav_calendar_todo(server, calendar_id, todo_id) as todo:
        if todo.icalendar_component.get('status', ''):
            del todo.icalendar_component['status']
            del todo.icalendar_component['completed']
            click.echo('Todo is pending')
        else:
            todo.icalendar_component.add('status', 'COMPLETED')
            todo.icalendar_component.add('completed', datetime.now(tz))
            click.echo('Todo completed')
        todo.save()


@click.command('percentage', options_metavar='[options]')
@click.argument('todo_id', metavar='<uid>')
@click.argument('percent', type=click.IntRange(0, 100), metavar='<int>')
@click.pass_obj
def percentage_complete(context, todo_id, percent):
    '''Set percentage completed'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    tz = context['config'].tz()

    with dav.caldav_calendar_todo(server, calendar_id, todo_id) as todo:
        if percent == todo.icalendar_component.get('percent-complete', 0):
            click.echo('No change!')
            return
        if 'percent-complete' in todo.icalendar_component:
            del todo.icalendar_component['percent-complete']
        todo.icalendar_component.add('percent-complete', percent)
        if todo.icalendar_component.get('status', ''):
            del todo.icalendar_component['status']
            del todo.icalendar_component['completed']
        if percent == 100:
            todo.icalendar_component.add('status', 'COMPLETED')
            todo.icalendar_component.add('completed', datetime.now(tz))
            click.echo('Todo completed')
        else:
            click.echo('Percentage set')
        todo.save()



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

import click
import icalendar
from caldav.lib import error

from caldavctl import dav, get_name, get_version

# Commands:

@click.command('backup')
@click.option('-f', '--file', type=click.File('w'), default=sys.stdout, help='Output file')
@click.pass_obj
def backup_calendar(context, file):
    '''
    Backup calendar to iCalendar file (ics)
    '''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    calendar_bck = icalendar.Calendar()
    calendar_bck.add('version', '2.0')
    calendar_bck.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')

    with dav.caldav_calendar(server, calendar_id) as calendar:
        for obj in calendar.objects(load_objects=True):
            # Parse the event's data into a Calendar object
            event_ical = icalendar.Calendar.from_ical(obj.icalendar_instance.to_ical())
            for component in event_ical.walk():
                if component.name in ('VEVENT', 'VTODO', 'VJOURNAL'):  # Filter desired components
                    calendar_bck.add_component(component)

    file.write(str(calendar_bck.to_ical(), 'utf-8'))


@click.command('restore')
@click.option('-f', '--file', type=click.File('r'), default=sys.stdin, help='Input file')
@click.pass_obj
def restore_calendar(context, file):
    '''
    Restore calendar to iCalendar file (ics)
    '''
    # Read the ICS file
    ics_content = file.read()

    # Parse the ICS file
    calendar_rst = icalendar.Calendar.from_ical(ics_content)

    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        for component in calendar_rst.walk():
            if component.name in ('VEVENT', 'VTODO', 'VJOURNAL'):
                obj = icalendar.Calendar()
                obj.add('version', '2.0')
                obj.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')
                obj.add_component(component)
                obj_string = obj.to_ical().decode('utf-8')
                try:
                    if component.name == 'VEVENT':
                        calendar.save_event(obj_string)
                    elif component.name == 'VTODO':
                        calendar.save_todo(obj_string)
                    elif component.name == 'VJOURNAL':
                        calendar.save_journal(obj_string)
                except error.AuthorizationError:
                    print(f'Error restoring object with uid {component["uid"]}. Skipping.')

    print('Restored.')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later


from caldavctl.click_utils import OptionsCompatibility


def test_click_utils_compatibility():
    compatibility = OptionsCompatibility(['today', 'week', 'dtstart', 'dtend'])
    compatibility.set_exception(['dtstart', 'dtend'])

    test_compatibility = {
        'today': {
            'week': False,
            'dtstart': False,
            'dtend': False,
        },
        'week': {
            'today': False,
            'dtstart': False,
            'dtend': False,
        },
        'dtstart': {
            'today': False,
            'week': False,
            'dtend': True,
        },
        'dtend': {
            'today': False,
            'week': False,
            'dtstart': True,
        },
    }

    assert test_compatibility == compatibility.compatibility



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
Configuration Management
'''

import os
import os.path
import tomllib
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import click
from platformdirs import user_config_dir

from caldavctl.utils import deep_merge_dict


class Config:
    def __init__(self, config_file, name=None, username=None, password=None,
                 url=None, timezone=None, default_server=None,
                 default_calendar=None):
        # Load the configuration file
        self.config = self.load_config_file(config_file)
        # Set the timezone:
        tz_name = self.config['default']['timezone']
        if isinstance(tz_name, str):
            try:
                self.config['default']['timezone'] = ZoneInfo(tz_name)
            except ZoneInfoNotFoundError:
                raise click.UsageError(f'Unknown timezone "{tz_name}"')
        # Add a server from the command line or environment variables if it exists
        self.merge_server(name, username, password, url)
        # Merge defaults from the command line if they exist
        self.merge_defaults(default_server, default_calendar, timezone)

    def check_file_permission(self, config_file):
        '''Check if the configuration file has 400 or 600 permissions'''
        file_perm = os.stat(config_file).st_mode & 0o777
        if file_perm != 0o400 and file_perm != 0o600:
            raise click.UsageError('The configuration file must be readable only by the user.')

    def validate_config(self, config):
        def check_allowed_keys(allowed, real):
            if not real.issubset(allowed):
                unknown = ','.join(list(real - allowed))
                raise click.UsageError(f'Unknown configuration directive: {unknown}')

        # Check first level keys
        allowed_keys = {'server', 'default', 'option'}
        check_allowed_keys(allowed_keys, set(config.keys()))

        # Check defaults
        allowed_defaults = {'server', 'calendar', 'timezone'}
        if 'default' in config:
            check_allowed_keys(allowed_defaults, set(config['default'].keys()))

        # Check servers:
        allowed_server = {'username', 'password', 'url'}
        if 'server' in config:
            for server in config['server']:
                real = set(config['server'][server])
                check_allowed_keys(allowed_server, real)
                if allowed_server - real:
                    missing = ', '.join(list(allowed_server - real))
                    raise click.UsageError(f'Missing configuration keys: {missing}')

    def load_config_file(self, config_file):
        config = {
            'default': {
                'calendar': None,
                'server': None,
                'timezone': ZoneInfo('UTC'),
            },
            'server': {},
            'option': []
        }
        if not config_file:
            # If no configuration file is declared,
            # use the default ~/.config/caldavctl/config.toml
            # in most systems
            config_dir = user_config_dir('caldavctl')
            config_file = os.path.join(config_dir, 'config.toml')

        if os.path.exists(config_file):
            self.check_file_permission(config_file)

            # Read the configuration file
            try:
                with open(config_file, 'rb') as f:
                    conf = tomllib.load(f)
                self.validate_config(conf)
                deep_merge_dict(conf, config)
            except tomllib.TOMLDecodeError as error:
                raise click.UsageError(f'Invalid configuration file: {error}')

        return config

    def merge_server(self, name, username, password, url):
        '''Merge a server with the configuration'''
        if name or username:
            name = name if name else 'default'
            self.config['server'][name] = {
                'username': username,
                'password': password,
                'url': url
            }

    def merge_defaults(self, server, calendar, timezone):
        if server:
            self.config['default']['server'] = server
        if calendar:
            self.config['default']['calendar'] = calendar
        if timezone:
            try:
                self.config['default']['timezone'] = ZoneInfo(timezone)
            except ZoneInfoNotFoundError:
                raise click.UsageError(f'Unknown timezone "{timezone}"')

    def tz(self):
        return self.config['default']['timezone']

    def servers(self):
        '''Iter all the servers'''
        for server_name in self.config['server']:
            yield self.get_server(server_name)

    def get_server(self, server_name=None):
        '''Get the default server or server_name'''
        if not server_name:
            try:
                server_name = self.config['default']['server']
                if not server_name:
                    raise click.UsageError('There is no default server defined and you did not '
                                           'indicate one with the --server option.')
            except KeyError:
                raise click.UsageError('There is no default server defined and you did not '
                                       'indicate one with the --server option.')
        server = self.config['server'][server_name]
        return server_name, server

    def get_calendar(self, calendar=None):
        if not calendar:
            try:
                calendar = self.config['default']['calendar']
            except KeyError:
                raise click.UsageError('There is no default calendar defined and you did not '
                                       'indicate one with the --calendar option.')
        return calendar



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
The tests on this file need a testing server that allows us to create a
calendar and that supports managing todos.

Is you have such calendar server configured, then you can run these tests by
defining the environment variable "RUN_NOMOCK_TESTS=yes".

The tests will create a new calendar, do the tests and then destroy the created
calendar.
'''

import os
import uuid

import pytest
from click.testing import CliRunner

from caldavctl import cli

RUN_NOMOCK_TESTS = os.environ.get('RUN_NOMOCK_TESTS', 'NO').upper() == 'YES'

uid = ''  # todo uid used on the tests


@pytest.fixture(scope="module")
def test_calendar():
    calendar_id = str(uuid.uuid4())
    runner = CliRunner()

    # Setup the testing environment - Create Calendar
    result = runner.invoke(
        cli.cli_group,
        ["calendar", "create", "--cal-id", calendar_id, "TestCalendar"]
    )
    assert result.exit_code == 0, f"Failed to create test calendar: {result.output}"

    yield calendar_id

    # Delete Calendar
    result = runner.invoke(
        cli.cli_group,
        ["calendar", "delete", calendar_id]
    )
    assert result.exit_code == 0, f"Failed to delete test calendar: {result.output}"


@pytest.fixture(scope="module")
def test_context():
    return {}


# Todo tests

@pytest.mark.skipif(not RUN_NOMOCK_TESTS, reason='RUN_NOMOCK_TESTS not set to "yes"')
def test_nomock_01_todo_create(test_calendar):
    calendar_id = test_calendar
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "create", "This is an example TODO"]
    )
    assert "todo created successfully:" in result.output


@pytest.mark.skipif(not RUN_NOMOCK_TESTS, reason='RUN_NOMOCK_TESTS not set to "yes"')
def test_nomock_02_todo_list(test_calendar, test_context):
    calendar_id = test_calendar
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "list", "-si"]
    )
    test_context['uid'] = result.output.split()[-1]
    assert '[ ]   0% This is an example TODO' in result.output


@pytest.mark.skipif(not RUN_NOMOCK_TESTS, reason='RUN_NOMOCK_TESTS not set to "yes"')
def test_nomock_03_todo_toggle_completed(test_calendar, test_context):
    uid = test_context['uid']
    calendar_id = test_calendar
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "toggle", uid]
    )
    assert 'Todo completed' in result.output
    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "toggle", uid]
    )
    assert 'Todo is pending' in result.output


@pytest.mark.skipif(not RUN_NOMOCK_TESTS, reason='RUN_NOMOCK_TESTS not set to "yes"')
def test_nomock_04_todo_percentage(test_calendar, test_context):
    uid = test_context['uid']
    calendar_id = test_calendar
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "percentage", uid, "45"]
    )
    assert 'Percentage set' in result.output

    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "percentage", uid, "45"]
    )
    assert 'No change' in result.output

    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "percentage", uid, "100"]
    )
    assert 'Todo completed' in result.output


@pytest.mark.skipif(not RUN_NOMOCK_TESTS, reason='RUN_NOMOCK_TESTS not set to "yes"')
def test_nomock_05_todo_delete(test_calendar, test_context):
    uid = test_context['uid']
    calendar_id = test_calendar
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ["--calendar", calendar_id, "todo", "delete", uid]
    )
    assert 'Todo deleted' in result.output



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import caldav
import click
from caldav.lib import error

from caldavctl import dav


# CALENDARS

@click.command('list', options_metavar='[options]')
@click.pass_obj
def list_calendars(context):
    '''
    List the available calendars present on the server or servers
    '''
    for name, server in context['config'].servers():
        with caldav.DAVClient(**server) as client:
            for calendar in client.principal().calendars():
                click.echo(f'Server {name}:')
                click.echo(f'    CALENDAR = {calendar.name}')
                click.echo(f'    ID = {calendar.id}')
                click.echo(f'    COMPONENTS = {', '.join(calendar.get_supported_components())}')
                click.echo(f'    URL = {calendar.url}')
                click.echo()


@click.command('create', options_metavar='[options]')
@click.argument('name', metavar='<calendar>')
@click.option('--cal-id', help='Calendar UID to use in the new calendar', metavar='<uid>')
@click.pass_obj
def create_calendar(context, name, cal_id=None):
    '''
    Create a calendar on the default server or optionally in another server

    <calendar> - calendar name
    '''
    _, server = context['config'].get_server()

    with caldav.DAVClient(**server) as client:
        principal = client.principal()
        try:
            new_calendar = principal.make_calendar(name=name, cal_id=cal_id)
        except error.AuthorizationError as msg:
            raise click.UsageError(f'Error creating the calendar (maybe duplicate UID?) with: {msg}')

        print(f'Calendar "{name}" created.')
    return new_calendar


@click.command('delete', options_metavar='[options]')
@click.argument('calendar-id', metavar='<uid>')
@click.pass_obj
def delete_calendar(context, calendar_id):
    '''
    Delete a calendar from the default server or optionally from another
    server. It's possible to have calendars with the same name, so we use the
    id to identify the calendar to delete.
    '''
    _, server = context['config'].get_server()

    with dav.caldav_calendar(server, calendar_id) as calendar:
        name = calendar.name
        calendar.delete()
        print(f'Calendar "{name}" deleted.')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
from unittest.mock import MagicMock, patch
from click.testing import CliRunner
from caldavctl.calendar_dav import list_calendars, create_calendar, delete_calendar


@pytest.fixture
def mock_context():
    """Fixture for the CLI context."""
    return {
        "config": MagicMock(
            servers=MagicMock(return_value=[("default", {"url": "testserver.com"})]),
            get_server=MagicMock(return_value=("default", {"url": "testserver.com"}))
        )
    }


@patch("caldav.DAVClient")
def test_list_calendars(mock_davclient, mock_context):
    """Test listing calendars."""
    # Mock principal and calendars
    mock_principal = MagicMock()

    class mock_calendar:
        name = "Test Calendar"
        id = "123"
        url = "http://testserver.com/calendar/123"
        get_supported_components = MagicMock(return_value=["VEVENT", "VTODO"])

    mock_principal.calendars.return_value = [mock_calendar]
    mock_davclient.return_value.__enter__.return_value.principal.return_value = mock_principal

    runner = CliRunner()
    result = runner.invoke(list_calendars, obj=mock_context)

    assert result.exit_code == 0
    assert "Server default:" in result.output
    assert "CALENDAR = Test Calendar" in result.output
    assert "ID = 123" in result.output
    assert "COMPONENTS = VEVENT, VTODO" in result.output
    assert "URL = http://testserver.com/calendar/123" in result.output


@patch("caldav.DAVClient")
def test_create_calendar(mock_davclient, mock_context):
    """Test creating a calendar."""
    mock_principal = MagicMock()
    mock_davclient.return_value.__enter__.return_value.principal.return_value = mock_principal

    runner = CliRunner()
    result = runner.invoke(
        create_calendar,
        ["New Calendar"],
        obj=mock_context,
    )

    assert result.exit_code == 0
    assert 'Calendar "New Calendar" created.' in result.output
    mock_principal.make_calendar.assert_called_once_with(name="New Calendar", cal_id=None)


@patch("caldav.DAVClient")
@patch("caldavctl.dav.caldav_calendar")
def test_delete_calendar(mock_caldav_calendar, mock_davclient, mock_context):
    """Test deleting a calendar."""
    class mock_calendar:
        name = "Test Calendar"
        delete = MagicMock()

    mock_caldav_calendar.return_value.__enter__.return_value = mock_calendar

    runner = CliRunner()
    result = runner.invoke(
        delete_calendar,
        ["bdcbe606-b68c-11ef-b03f-d4f32d4b839a"],
        obj=mock_context,
    )

    assert result.exit_code == 0
    assert 'Calendar "Test Calendar" deleted.' in result.output
    mock_caldav_calendar.assert_called_once_with({"url": "testserver.com"},
                                                 "bdcbe606-b68c-11ef-b03f-d4f32d4b839a")
    mock_calendar.delete.assert_called_once()



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime, timedelta

import click

from caldavctl import dav
from caldavctl.click_utils import create_compatibility_check
from caldavctl.utils import edit_text_with_editor, to_datetime

exclusive_option = create_compatibility_check(
    ['today', 'week', 'dtstart', 'dtend'],
    [['dtstart', 'dtend']])


@click.command('list')
@click.option('-t', '--today',
              is_flag=True, show_default=True, default=False,
              callback=exclusive_option, help='Today\'s journals',)
@click.option('-w', '--week',
              is_flag=True, show_default=True, default=False,
              callback=exclusive_option, help='This week\'s journals')
@click.option('-s', '--dtstart',
              callback=exclusive_option, help='Date range, start date')
@click.option('-e', '--dtend',
              callback=exclusive_option, help='Date range, end date')
@click.option('-d', '--show_description',
              is_flag=True, show_default=True, default=False,
              help='Show the journal\'s description.')
@click.pass_obj
def list_journals(context, today, week, dtstart, dtend, show_description):
    '''
    List journals from the server list.
    '''
    raise click.UsageError('caldav does not support journal searching. '
                           'https://github.com/python-caldav/caldav/issues/237')

    tz = context['config'].tz()
    start_date = None
    end_date = None

    if today:
        start_date = datetime.now().replace(hour=0, minute=0, second=0)
        end_date = datetime.now().replace(hour=23, minute=59, second=59)
    elif week:
        today_date = datetime.now().replace(hour=0, minute=0, second=0)
        start_date = today_date - timedelta(days=today_date.weekday())
        end_date = start_date + timedelta(days=6)
        end_date = end_date.replace(hour=23, minute=59, second=59)
    elif dtstart and dtend:
        if dtstart:
            start_date = to_datetime(dtstart, tz)
        else:
            start_date = datetime(1900, 1, 1, tzinfo=tz)
        if dtend:
            end_date = to_datetime(dtend, tz)
        else:
            end_date = datetime(5000, 1, 1, tzinfo=tz)
    else:
        raise click.UsageError('You have to specify a time frame, for example with -t or -w.')

    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        # Check if the calendar supports journals:
        if 'VJOURNAL' not in calendar.get_supported_components():
            raise click.UsageError('This calendar does not support journals.')

        # Get journals in time range
        journals = calendar.search(
            start=start_date,
            end=end_date,
            journal=True,
            expand=True,
            sort_keys=['dtstart']
        )
        for journal in journals:
            ev = journal.icalendar_component
            date = ev.get('dtstart').dt
            summary = ev.get('summary', '')
            description = ev.get('description', '')
            uid = ev.get('uid')

            if isinstance(date, datetime):
                print(f'{date.strftime('%Y-%m-%d %H:%M')} - {summary} - {uid}')
            else:
                print(f'{date} (day) - {summary} - {uid}')
            if show_description and description.strip():
                print(description)



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from contextlib import contextmanager

import caldav
import caldav.elements.ical
import click
from caldav.lib.error import AuthorizationError
from requests import ConnectionError


@contextmanager
def caldav_calendar(server, calendar_id):
    '''
    Open a connection to a server and return the calendar object
    '''
    # Code to acquire resource, e.g.:
    client = caldav.DAVClient(**server)

    # Retrieve all calendars for the principal
    try:
        principal = client.principal()
    except ConnectionError:
        raise click.UsageError('Connection error, check your server url')
    except AuthorizationError:
        raise click.UsageError('Authorization error, check your user name and password')
    calendars = principal.calendars()

    # Find the calendar by URL or iterate through all calendars
    for cal in calendars:
        if cal.id == calendar_id:  # Match the desired calendar URL
            calendar = cal
            break
    else:
        raise click.UsageError(f'Calendar with id "{calendar_id}" not found')

    # Yield the calendar and close it if appropriate
    try:
        yield calendar
    finally:
        client.close()


@contextmanager
def caldav_calendar_todo(server, calendar_id, todo_id):
    '''
    Open a connection to a server, select a calendar and return a todo object
    '''
    with caldav_calendar(server, calendar_id) as calendar:
        try:
            todo = calendar.todo_by_uid(todo_id)
        except caldav.lib.error.NotFoundError:
            raise click.UsageError(f'Could not find todo {todo_id}.')
        yield todo


@contextmanager
def caldav_calendar_event(server, calendar_id, event_id):
    '''
    Open a connection to a server, select a calendar and return a event object
    '''
    with caldav_calendar(server, calendar_id) as calendar:
        try:
            event = calendar.event_by_uid(event_id)
        except caldav.lib.error.NotFoundError:
            raise click.UsageError(f'Could not find event {event_id}.')
        yield event


@contextmanager
def caldav_calendar_journal(server, calendar_id, journal_id):
    '''
    Open a connection to a server, select a calendar and return a journal object
    '''
    with caldav_calendar(server, calendar_id) as calendar:
        try:
            journal = calendar.journal_by_uid(journal_id)
        except caldav.lib.error.NotFoundError:
            raise click.UsageError(f'Could not find journal {journal_id}.')
        yield journal



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from itertools import combinations


class OptionsCompatibility:
    def __init__(self, option_list):
        '''
        The `option_list` is a list of options that can't be used together.
        '''
        self.option_list = list(set(option_list))
        self.set_compatibility()

    def set_compatibility(self):
        incomp = {}
        for option in self.option_list:
            incomp[option] = {opt: False for opt in self.option_list if opt != option}
        self.compatibility = incomp

    def set_exception(self, options):
        for opt1, opt2 in combinations(options, 2):
            self.compatibility[opt1][opt2] = True
            self.compatibility[opt2][opt1] = True

    def check(self, opt1, opt2):
        return self.compatibility[opt1][opt2]


def create_compatibility_check(options, exceptions):
    '''
    options - list of options that are incompatible
    exclusions - list of exclusions, it's a list of lists
    '''
    compatibility = OptionsCompatibility(options)
    for exception in exceptions:
        compatibility.set_exception(exception)

    def exclusive_option(context, param, value):
        if value:
            context.obj['option'].append(param.name)
        if len(context.obj['option']) > 1:
            for opt1, opt2 in combinations(context.obj['option'], 2):
                if not compatibility.check(opt1, opt2):
                    context.exit('Error, the options '
                                 f'"{opt1}" and "{opt2}" can\'t be used together.')
        return value

    return exclusive_option



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from zoneinfo import ZoneInfo

import pytest

from click.testing import CliRunner
from unittest.mock import patch, MagicMock
from caldavctl.todo import (
    list_todos, create_todo, delete_todo, toggle_todo_complete, percentage_complete
)


@pytest.fixture
def mock_context():
    """Fixture for the CLI context."""
    return {
        "config": MagicMock(
            servers=MagicMock(return_value=[("default", {"url": "testserver.com"})]),
            get_server=MagicMock(return_value=("default", {"url": "testserver.com"})),
            tz=MagicMock(return_value=ZoneInfo('Europe/Lisbon'))
        )
    }


@patch("caldavctl.dav.caldav_calendar")
def test_list_todos(mock_caldav_calendar, mock_context):

    class mock_calendar:
        name = "Test Calendar"
        id = "123"
        url = "http://testserver.com/calendar/123"
        get_supported_components = MagicMock(return_value=["VEVENT", "VTODO"])
        todos = MagicMock(return_value=[
            MagicMock(icalendar_component={
                'summary': 'Test Todo',
                'status': '',
                'percent-complete': 50,
                'description': 'Test Description',
                'uid': '1234'
                })
        ])

    mock_caldav_calendar.return_value.__enter__.return_value = mock_calendar()

    runner = CliRunner()
    result = runner.invoke(list_todos, ['--show-description', '--show-uid'], obj=mock_context)

    assert result.exit_code == 0
    assert '[ ]  50% Test Todo - 1234' in result.output
    assert 'Test Description' in result.output


@patch("caldavctl.dav.caldav_calendar")
def test_create_todo(mock_caldav_calendar, mock_context):
    runner = CliRunner()
    mock_calendar = MagicMock()
    mock_caldav_calendar.return_value.__enter__.return_value = mock_calendar

    result = runner.invoke(
        create_todo, [
            'Test Todo',
            '--description', 'Test Description',
            '--due-date', '2024-12-31T23:59:59',
            '--priority', '1'
        ],
        obj=mock_context
    )

    assert result.exit_code == 0
    assert 'todo created successfully: Test Todo' in result.output
    assert mock_calendar.save_event.called


@patch("caldavctl.dav.caldav_calendar_todo")
def test_delete_todo(mock_caldav_calendar_todo, mock_context):
    runner = CliRunner()
    mock_todo = MagicMock()

    mock_caldav_calendar_todo.return_value.__enter__.return_value = mock_todo
    result = runner.invoke(delete_todo, ['1234'], obj=mock_context)

    assert result.exit_code == 0
    assert 'Todo deleted' in result.output
    assert mock_todo.delete.called


@patch("caldavctl.dav.caldav_calendar_todo")
def test_toggle_todo_complete(mock_caldav_calendar_todo, mock_context):
    runner = CliRunner()
    mock_todo = MagicMock()

    # Test marking as completed
    mock_todo.icalendar_component.get.return_value = ''

    mock_caldav_calendar_todo.return_value.__enter__.return_value = mock_todo
    result = runner.invoke(toggle_todo_complete, ['1234'], obj=mock_context)

    assert result.exit_code == 0
    assert 'Todo completed' in result.output
    mock_todo.save.assert_called()

    # Test marking as not completed
    mock_todo.icalendar_component.get.return_value = 'COMPLETED'

    mock_caldav_calendar_todo.return_value.__enter__.return_value = mock_todo
    result = runner.invoke(toggle_todo_complete, ['1234'], obj=mock_context)

    assert result.exit_code == 0
    assert 'Todo is pending' in result.output
    mock_todo.save.assert_called()


@patch("caldavctl.dav.caldav_calendar_todo")
def test_percentage_complete(mock_caldav_calendar_todo, mock_context):
    runner = CliRunner()
    mock_todo = MagicMock()
    mock_todo.icalendar_component.get.side_effect = lambda key, default: {
        'percent-complete': 50
    }.get(key, default)

    mock_caldav_calendar_todo.return_value.__enter__.return_value = mock_todo
    result = runner.invoke(percentage_complete, ['1234', '75'], obj=mock_context)

    assert result.exit_code == 0
    assert 'Percentage set' in result.output
    mock_todo.save.assert_called()



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import uuid
from datetime import datetime

from icalendar import Alarm, Calendar, Event

from caldavctl import get_name, get_version
from caldavctl.object_parser import (
    DATE,
    INTEGER,
    LIST,
    STRING,
    TZ,
    ObjectParser,
    ParserError,
    ValueParser,
)
from caldavctl.utils import check_if_naive, duration_validation, duration_to_timedelta


class AlarmParser(ValueParser):
    def tokenize(self):
        self.tokens = self.value.split(maxsplit=1)

    def validate(self):
        duration = self.tokens[0].upper().strip()
        if not duration_validation(duration):
            raise ParserError(f'Invalid duration "{self.tokens[0]}" for alarm trigger.')
        self.tokens[0] = duration_to_timedelta(duration)


EVENT_LEXICON = {
    # token: (type, optional?)
    'summary': (STRING, True),
    'location': (STRING, True),
    'categories': (LIST, True),
    'dtstart': (DATE, False),
    'dtend': (DATE, False),
    'description': (STRING, True),
    'percentage': (INTEGER, True),
    'timezone': (TZ, True),
    'priority': (INTEGER, True),
    'alarm': (AlarmParser, True)
}


def parse_event(event_icalendar, timezone):
    result = ObjectParser(event_icalendar, EVENT_LEXICON).run()

    # Check if the start and/or end dates are naïve
    tz = timezone if 'timezone' not in result else result['timezone']
    result['dtstart'] = check_if_naive(result['dtstart'], tz)
    result['dtend'] = check_if_naive(result['dtend'], tz)

    return result


event_optional_keys = (
    'summary',
    'location',
    'priority',
    'description',
    'categories',
    'alarm',
)


def event_builder(event_data, tz):
    calendar = Calendar()
    calendar.add('version', '2.0')
    calendar.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')

    event = Event()
    # Mandatory keys in a VEVENT object
    event.add('uid', str(uuid.uuid4()))
    event.add('dtstamp', datetime.now(tz))
    event.add('dtstart', event_data.get('dtstart'))
    event.add('dtend', event_data.get('dtend'))
    # Optional keys:
    for key in event_optional_keys:
        if key in event_data:
            if key == 'categories':
                for category in event_data.get('categories'):
                    event.add('categories', category)
            if key == 'alarm':
                data = event_data.get(key)
                for alarm_data in data:
                    alarm = Alarm()
                    alarm.add('trigger', alarm_data[0])
                    if len(alarm_data) == 2:
                        alarm.add('description', alarm_data[1])
                    else:
                        alarm.add('description', 'Default caldavctl description')
                    alarm.add('action', 'display')
                    event.add_component(alarm)
            else:
                event.add(key, event_data.get(key))

    calendar.add_component(event)

    return calendar.to_ical().decode('utf-8')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import uuid

import click
import icalendar

from caldavctl import dav


@click.command('get-todo', options_metavar='[options]')
@click.argument('todo_uid', metavar='<uid>')
@click.option('-f', '--file', type=click.File('w'), default=sys.stdout, metavar='<file>')
@click.pass_obj
def get_todo(context, todo_uid, file):
    '''Get the ICS file of a todo'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_todo(server, calendar_id, todo_uid) as todo:
        file.write(str(todo.icalendar_instance.to_ical(), 'utf-8'))


@click.command('create-todo', options_metavar='[options]')
@click.argument('file', type=click.File('r'), default=sys.stdin, metavar='<file>')
@click.pass_obj
def create_todo(context, file):
    '''
    Create a todo from an existing ICS File

    If the ICS already has a UID, and if a todo already exists on the server
    with that ICS, them the server copy will be modified with the file
    contents.
    '''
    # Read the ICS file
    ics_content = file.read()

    # Parse the ICS file
    cal = icalendar.Calendar.from_ical(ics_content)

    # Find the first Todo component in the file
    todo = None
    for component in cal.walk():
        if component.name == 'VTODO':
            todo = component
            break

    if not todo:
        raise click.UsageError('No Todo found in the ICS file')

    # Ensure the todo has a unique identifier
    if 'uid' not in todo:
        todo['uid'] = str(uuid.uuid4())

    # Convert the todo back to a string
    todo_string = todo.to_ical().decode('utf-8')

    # Save the todo to the calendar
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    with dav.caldav_calendar(server, calendar_id) as calendar:
        calendar.save_event(todo_string)

    click.echo('Todo created successfully')


@click.command('get-event', options_metavar='[options]')
@click.argument('event_uid', metavar='<uid>')
@click.option('-f', '--file', type=click.File('w'), default=sys.stdout, help='Output file', metavar='<file>')
@click.pass_obj
def get_event(context, event_uid, file):
    '''Get the ICS file of a event'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_event(server, calendar_id, event_uid) as event:
        file.write(str(event.icalendar_instance.to_ical(), 'utf-8'))



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later


import os
import re
import subprocess
import tempfile

from datetime import datetime, timedelta

import click


def edit_text_with_editor(initial_text: str, suffix: str = '.tmp') -> str:
    # Determine the editor from $EDITOR or $VISUAL, defaulting to 'vi'
    editor = os.environ.get('VISUAL') or os.environ.get('EDITOR') or 'vi'

    # Create a temporary file
    with tempfile.NamedTemporaryFile(suffix=suffix, delete=False, mode='w+') as temp_file:
        temp_filename = temp_file.name
        temp_file.write(initial_text)
        temp_file.flush()  # Ensure content is written to disk

    try:
        # Open the editor with the temporary file
        subprocess.run([editor, temp_filename])

        # Read the modified content back
        with open(temp_filename, 'r') as temp_file:
            modified_text = temp_file.read()
    finally:
        # Clean up the temporary file
        os.unlink(temp_filename)

    return modified_text


def deep_merge_dict(source, destination):
    """
    Merge two dictionaries, if key exists on both dictionaries then the source will prevail

    Based on:
    From: https://stackoverflow.com/questions/20656135/python-deep-merge-dictionary-data
    """
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            if isinstance(node, dict):
                # see test_deep_merge_dict_conflicting_types
                deep_merge_dict(value, node)
            else:
                destination[key] = value
        else:
            destination[key] = value
    return destination


def check_if_naive(date, tz):
    '''
    Check if a date is naïve and if it is localize it with the timezone defined
    in tz.
    '''
    if date.tzinfo is None or date.tzinfo.utcoffset(date) is None:
        date = date.replace(tzinfo=tz)
    return date


def to_datetime(dt_str, tz):
    try:
        dt = datetime.fromisoformat(dt_str)
    except ValueError:
        raise click.UsageError(f'Invalid date "{dt_str}", use iso format.')
    return dt


def duration_validation(duration):
    '''
    Validates a duration difined according to RFC 2445 iCalendar specification:

    dur-value  = (["+"] / "-") "P" (dur-date / dur-time / dur-week)
    dur-date   = dur-day [dur-time]
    dur-time   = "T" (dur-hour / dur-minute / dur-second)
    dur-week   = 1*DIGIT "W"
    dur-hour   = 1*DIGIT "H" [dur-minute]
    dur-minute = 1*DIGIT "M" [dur-second]
    dur-second = 1*DIGIT "S"
    dur-day    = 1*DIGIT "D"
    '''

    h = r'\d+H'
    m = r'\d+M'
    s = r'\d+S'
    dur_time = rf'T(?:{h}{m}{s}|{h}{m}|{h}{s}|{m}{s}|{h}|{m}|{s})'

    duration_re = (
        r'^[+-]?P'
        r'(?:'
        r'\d+W|'
        rf'\d+D(?:{dur_time})?|'
        rf'{dur_time}'
        r')$'
    )

    duration = duration.upper().strip()
    return bool(re.match(duration_re, duration))


def duration_to_timedelta(duration):
    '''
    Converts durations used in iCalendar files to python's timedelta

    No attempt is made to validade the "duration" format. If the "duration" is
    malformed the result is undefined.
    '''
    translate = {
        'W': 'weeks',
        'D': 'days',
        'H': 'hours',
        'M': 'minutes',
        'S': 'seconds'
    }

    duration = duration.upper().strip()
    signal = 1
    tmp_int = ''
    duration_dict = {}
    for i, ch in enumerate(duration):
        if i == 0 and ch == '-':
            signal = -1
        elif ch.isdigit():
            tmp_int += ch
        elif ch in ('P', 'T'):
            pass
        else:
            duration_dict[translate[ch]] = int(tmp_int)
            tmp_int = ''

    return signal * timedelta(**duration_dict)



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@tretas.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import uuid
from datetime import datetime

from icalendar import Calendar, Journal

from caldavctl import get_name, get_version
from caldavctl.object_parser import DATE, INTEGER, LIST, STRING, TZ, ObjectParser
from caldavctl.utils import check_if_naive

JOURNAL_LEXICON = {
    # token: (type, optional?)
    'summary': (STRING, True),
    'categories': (LIST, True),
    'dtstart': (DATE, False),
    'description': (STRING, True),
    'percentage': (INTEGER, True),
    'timezone': (TZ, True),
    'priority': (INTEGER, True),
}


def parse_journal(journal_icalendar, timezone):
    result = ObjectParser(journal_icalendar, JOURNAL_LEXICON).run()

    # Check if the start and/or end dates are naïve
    tz = timezone if 'timezone' not in result else result['timezone']
    result['dtstart'] = check_if_naive(result['dtstart'], tz)
    result['dtend'] = check_if_naive(result['dtend'], tz)

    return result


def journal_builder(journal_data, tz):
    journal_optional_keys = (
        'summary',
        'location',
        'priority',
        'description',
        'categories',
    )

    calendar = Calendar()
    calendar.add('version', '2.0')
    calendar.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')

    journal = Journal()
    # Mandatory keys in a VJOURNAL object
    journal.add('uid', str(uuid.uuid4()))
    journal.add('dtstamp', datetime.now(tz))
    journal.add('dtstart', journal_data.get('dtstart'))
    # Optional keys:
    for key in journal_optional_keys:
        if key in journal_data:
            if key == 'categories':
                for category in journal_data.get('categories'):
                    journal.add('categories', category)
            else:
                journal.add(key, journal_data.get(key))

    calendar.add_component(journal)

    return calendar.to_ical().decode('utf-8')
